# -*- coding: utf-8 -*-

import time
import os
import subprocess
import sys
import signal
from .request import Request
from .connection import Connection

from .mixins import RoutesMixin, AppEventsMixin
from ..log import logger
from .. import constants


class Worker(RoutesMixin, AppEventsMixin):
    connection_class = Connection
    request_class = Request

    box_class = None

    debug = False
    got_first_request = False
    blueprints = None
    # 是否有效(父进程中代表程序有效，子进程中代表worker是否有效)
    enable = True
    # 子进程列表
    processes = None
    # 最多回应一次
    rsp_once = True

    host = None
    port = None

    def __init__(self, box_class):
        RoutesMixin.__init__(self)
        AppEventsMixin.__init__(self)
        self.blueprints = list()
        self.processes = list()
        self.box_class = box_class

    def register_blueprint(self, blueprint):
        blueprint.register_to_app(self)

    def run(self, host=None, port=None, debug=None, workers=None, rsp_once=None):
        self.host = host
        self.port = port

        if debug is not None:
            self.debug = debug

        workers = workers if workers is not None else 1

        if rsp_once is not None:
            self.rsp_once = rsp_once

        if not os.getenv(constants.WORKER_ENV_KEY):
            # 主进程
            logger.info('Connect to server on %s:%s, debug: %s, workers: %s',
                        host, port, self.debug, workers)

            # 只能在主线程里面设置signals
            self._handle_parent_proc_signals()
            self._fork_workers(workers)
        else:
            # 子进程
            self._try_serve_forever()

    def _before_worker_run(self):
        self.events.create_worker()
        for bp in self.blueprints:
            bp.events.create_app_worker()

    def _try_serve_forever(self):
        self._handle_child_proc_signals()

        self._before_worker_run()

        try:
            self._serve_forever()
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur.', exc_info=True)

    def _fork_workers(self, workers):
        def start_worker_process():
            args = [sys.executable] + sys.argv
            inner_p = subprocess.Popen(args, env={
                constants.WORKER_ENV_KEY: '1'
            })
            return inner_p

        for it in xrange(0, workers):
            p = start_worker_process()
            self.processes.append(p)

        while 1:
            for idx, p in enumerate(self.processes):
                if p and p.poll() is not None:
                    # 说明退出了
                    self.processes[idx] = None

                    if self.enable:
                        # 如果还要继续服务
                        p = start_worker_process()
                        self.processes[idx] = p

            if not filter(lambda x: x, self.processes):
                # 没活着的了
                break

            # 时间短点，退出的快一些
            time.sleep(0.1)

    def _handle_parent_proc_signals(self):
        def exit_handler(*args, **kwargs):
            self.enable = False

            for p in self.processes:
                if p:
                    p.send_signal(signal.SIGTERM)

        def reload_handler(*args, **kwargs):
            """
            让所有子进程重新加载
            """
            for p in self.processes:
                if p:
                    p.send_signal(signal.SIGUSR1)

        signal.signal(signal.SIGTERM, exit_handler)
        signal.signal(signal.SIGINT, exit_handler)
        signal.signal(signal.SIGUSR1, reload_handler)

    def _handle_child_proc_signals(self):
        def reload_handler(*args, **kwargs):
            self.enable = False

        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGUSR1, reload_handler)

    def _serve_forever(self):
        conn = self.connection_class(self, self.host, self.port)
        conn.run()
